# -*- coding: utf-8 -*-
#
# This file is part of the StatsCrosslet project
#
#
#
# Distributed under the terms of the APACHE license.
# See LICENSE.txt for more info.

""" OPC-UA client for LOFAR stations crosslet stats

"""

# PyTango imports
import tango
from tango import DebugIt
from tango.server import run
from tango.server import Device
from tango.server import attribute, command
from tango.server import device_property
from tango import AttrQuality, DispLevel, DevState
from tango import AttrWriteType, PipeWriteType
# Additional import
# PROTECTED REGION ID(StatsCrosslet.additionnal_import) ENABLED START #
import opcua
import numpy
import traceback
import threading
from datetime import datetime
# PROTECTED REGION END #    //  StatsCrosslet.additionnal_import

__all__ = ["StatsCrosslet", "main"]


class StatsCrosslet(Device):
    """

    **Properties:**

    - Device Property
        OPC_Server_Name
            - Type:'DevString'
        OPC_Server_Port
            - Type:'DevULong'
        OPC_Time_out
            - Type:'DevDouble'
        Default_pause_time
            - Type:'DevDouble'
        Default_subband
            - Type:'DevULong'
        Default_integration_time
            - Type:'DevDouble'
    """
    # PROTECTED REGION ID(StatsCrosslet.class_variable) ENABLED START #
    client = 0
    ns = 0
    obj = 0
    record_cross = 0
    stop_data_read_loop = False
    data_acquisition_is_active = False
    data_read_loop = threading.Thread()

    def read_data(self):
        # This is the the thread that continuously reads the crosslet
        # statistics from the station and feeds it into the Tango
        # property system.
        self.debug_stream("Entering read_data loop.")
        while self.stop_data_read_loop is False:
            self.debug_stream("read_data is running...")
            threading.Event().wait(self._pause_time)
            if self.data_acquisition_is_active is True:
                try:
                    self.debug_stream("fetching data:  subband = %d, integration_time = %d", self._subband, self._integration_time)
                    t, visibilities_list, rcu_modes = self.obj.call_method(self.record_cross, self._subband, self._integration_time)
                    self.debug_stream("fetching data done: t = %s, len(visibilities_list) = %d, len(rcu_modes) = %d", t, len(visibilities_list), len(rcu_modes))
                    self._time_stamp = t
                    visibilities = numpy.array(visibilities_list)[0] + 1j * numpy.array(visibilities_list[1])
                    self._visibilities_real = visibilities.real
                    self._visibilities_imag = visibilities.imag
                    self._rcu_modes = rcu_modes
                    self.debug_stream("timestamp = %s, visibilities_real [0][1] = %f", self._time_stamp, self._visibilities_real[0][1])
                except KeyboardInterrupt:
                    self.debug_stream("KBD interrupt caught, leaving data read loop.")
                    self.stop_data_read_loop = True
                except Exception as e:
                    self.error_stream("Exception: %s.  Cannot call the method %s in the OPC-UA server %s.  Trace: %s", e, self.record_cross, self.OPC_Server_Name, traceback.format_exc())

    # PROTECTED REGION END #    //  StatsCrosslet.class_variable

    # -----------------
    # Device Properties
    # -----------------

    OPC_Server_Name = device_property(
        dtype='DevString',
        default_value="okeanos-kabel"
    )

    OPC_Server_Port = device_property(
        dtype='DevULong',
        default_value=55556
    )

    OPC_Time_out = device_property(
        dtype='DevDouble',
        default_value=1.0
    )

    Default_pause_time = device_property(
        dtype='DevDouble',
        default_value=1.0
    )

    Default_subband = device_property(
        dtype='DevULong',
        default_value=150
    )

    Default_integration_time = device_property(
        dtype='DevDouble',
        default_value=1.0
    )

    # ----------
    # Attributes
    # ----------

    Subband = attribute(
        dtype='DevUShort',
        access=AttrWriteType.READ_WRITE,
    )

    Integration_time = attribute(
        dtype='DevDouble',
    )

    Time_stamp = attribute(
        dtype='DevString',
    )

    Pause_time = attribute(
        dtype='DevDouble',
        access=AttrWriteType.READ_WRITE,
    )

    RCU_modes = attribute(
        dtype=('DevString',),
        max_dim_x=96,
    )

    Visibilities_imag = attribute(
        dtype=(('DevDouble',),),
        max_dim_x=96, max_dim_y=96,
    )

    Visibilities_real = attribute(
        dtype=(('DevDouble',),),
        max_dim_x=96, max_dim_y=96,
    )

    # ---------------
    # General methods
    # ---------------

    def init_device(self):
        """Initialises the attributes and properties of the StatsCrosslet."""
        Device.init_device(self)
        # PROTECTED REGION ID(StatsCrosslet.init_device) ENABLED START #
        self.set_state(DevState.INIT)
        # Set default values in the read/write attributes.
        self._time_stamp = ''
        self._rcu_modes = ('',)
        self._visibilities_imag = ((0.0,),)
        self._visibilities_real = ((0.0,),)

        # Set defaults to property values.
        self._subband = self.Default_subband
        self._integration_time = self.Default_integration_time
        self._pause_time = self.Default_pause_time

        try:
            self.debug_stream("Connecting to OPC-UA server %s:%d...", self.OPC_Server_Name, self.OPC_Server_Port)
            self.client = opcua.Client("opc.tcp://{}:{}/".format(self.OPC_Server_Name, self.OPC_Server_Port), self.OPC_Time_out * 1000)
            self.client.connect()
            ns = self.client.get_namespace_index("http://lofar.eu")
            self.obj = self.client.get_root_node().get_child(["0:Objects", "{}:StationMetrics".format(ns),                                            "{}:RCU".format(ns)])
            self.record_cross = "{}:record_cross".format(ns)
            self.debug_stream("Connecting to OPC-UA server %s:%d done.", self.OPC_Server_Name, self.OPC_Server_Port)

            self.data_read_loop = threading.Thread(target = self.read_data)
            self.data_acquisition_is_active = False
            self.stop_data_read_loop = False
            self.data_read_loop.start()
            self.set_state(DevState.ON)
        except Exception as e:
            self.set_state(DevState.FAULT)
            self.error_stream("Cannot connect to the OPC-UA server %s.  Trace: %s" % (self.OPC_Server_Name, traceback.format_exc()))
            raise e
        # PROTECTED REGION END #    //  StatsCrosslet.init_device

    def always_executed_hook(self):
        """Method always executed before any TANGO command is executed."""
        # PROTECTED REGION ID(StatsCrosslet.always_executed_hook) ENABLED START #
        # PROTECTED REGION END #    //  StatsCrosslet.always_executed_hook

    def delete_device(self):
        """Hook to delete resources allocated in init_device.

        This method allows for any memory or other resources allocated in the
        init_device method to be released.  This method is called by the device
        destructor and by the device Init command.
        """
        # PROTECTED REGION ID(StatsCrosslet.delete_device) ENABLED START #
        self.set_state(DevState.OFF)

        self.debug_stream("Shutting down...")
        self.data_acquisition_is_active = False
        self.stop_data_read_loop = True
        self.debug_stream("Waiting for data acquisition thread to shut down...")
        self.data_read_loop.join()
        self.debug_stream("Waiting for data acquisition thread to shut down, done")
        if self.client is not None:
            self.client.close_session()
            self.client.close_secure_channel()
        self.debug_stream("Shutdown done.")
        # PROTECTED REGION END #    //  StatsCrosslet.delete_device
    # ------------------
    # Attributes methods
    # ------------------

    def read_Subband(self):
        # PROTECTED REGION ID(StatsCrosslet.Subband_read) ENABLED START #
        """Return the Subband attribute."""
        return self.__subband
        # PROTECTED REGION END #    //  StatsCrosslet.Subband_read

    def write_Subband(self, value):
        # PROTECTED REGION ID(StatsCrosslet.Subband_write) ENABLED START #
        """Set the Subband attribute."""
        pass
        # PROTECTED REGION END #    //  StatsCrosslet.Subband_write

    def is_Subband_allowed(self, attr):
        # PROTECTED REGION ID(StatsCrosslet.is_Subband_allowed) ENABLED START #
        if attr==attr.READ_REQ:
            return self.get_state() not in [DevState.STANDBY]
        else:
            return self.get_state() not in [DevState.STANDBY]
        # PROTECTED REGION END #    //  StatsCrosslet.is_Subband_allowed

    def read_Integration_time(self):
        # PROTECTED REGION ID(StatsCrosslet.Integration_time_read) ENABLED START #
        """Return the Integration_time attribute."""
        return self.__integration_time
        # PROTECTED REGION END #    //  StatsCrosslet.Integration_time_read

    def is_Integration_time_allowed(self, attr):
        # PROTECTED REGION ID(StatsCrosslet.is_Integration_time_allowed) ENABLED START #
        return self.get_state() not in [DevState.ON,DevState.OFF,DevState.INIT]
        # PROTECTED REGION END #    //  StatsCrosslet.is_Integration_time_allowed

    def read_Time_stamp(self):
        # PROTECTED REGION ID(StatsCrosslet.Time_stamp_read) ENABLED START #
        """Return the Time_stamp attribute."""
        return self.__time_stamp
        # PROTECTED REGION END #    //  StatsCrosslet.Time_stamp_read

    def is_Time_stamp_allowed(self, attr):
        # PROTECTED REGION ID(StatsCrosslet.is_Time_stamp_allowed) ENABLED START #
        return self.get_state() not in [DevState.ON,DevState.OFF,DevState.INIT]
        # PROTECTED REGION END #    //  StatsCrosslet.is_Time_stamp_allowed

    def read_Pause_time(self):
        # PROTECTED REGION ID(StatsCrosslet.Pause_time_read) ENABLED START #
        """Return the Pause_time attribute."""
        return self.__pause_time
        # PROTECTED REGION END #    //  StatsCrosslet.Pause_time_read

    def write_Pause_time(self, value):
        # PROTECTED REGION ID(StatsCrosslet.Pause_time_write) ENABLED START #
        """Set the Pause_time attribute."""
        pass
        # PROTECTED REGION END #    //  StatsCrosslet.Pause_time_write

    def read_RCU_modes(self):
        # PROTECTED REGION ID(StatsCrosslet.RCU_modes_read) ENABLED START #
        """Return the RCU_modes attribute."""
        return self._rcu_modes
        # PROTECTED REGION END #    //  StatsCrosslet.RCU_modes_read

    def is_RCU_modes_allowed(self, attr):
        # PROTECTED REGION ID(StatsCrosslet.is_RCU_modes_allowed) ENABLED START #
        return self.get_state() not in [DevState.ON,DevState.OFF,DevState.INIT]
        # PROTECTED REGION END #    //  StatsCrosslet.is_RCU_modes_allowed

    def read_Visibilities_imag(self):
        # PROTECTED REGION ID(StatsCrosslet.Visibilities_imag_read) ENABLED START #
        """Return the Visibilities_imag attribute."""
        return self.__visibilities_imag
        # PROTECTED REGION END #    //  StatsCrosslet.Visibilities_imag_read

    def is_Visibilities_imag_allowed(self, attr):
        # PROTECTED REGION ID(StatsCrosslet.is_Visibilities_imag_allowed) ENABLED START #
        return self.get_state() not in [DevState.ON,DevState.OFF,DevState.INIT]
        # PROTECTED REGION END #    //  StatsCrosslet.is_Visibilities_imag_allowed

    def read_Visibilities_real(self):
        # PROTECTED REGION ID(StatsCrosslet.Visibilities_real_read) ENABLED START #
        """Return the Visibilities_real attribute."""
        return self.__visibilities_real
        # PROTECTED REGION END #    //  StatsCrosslet.Visibilities_real_read

    def is_Visibilities_real_allowed(self, attr):
        # PROTECTED REGION ID(StatsCrosslet.is_Visibilities_real_allowed) ENABLED START #
        return self.get_state() not in [DevState.ON,DevState.OFF,DevState.INIT]
        # PROTECTED REGION END #    //  StatsCrosslet.is_Visibilities_real_allowed

    # --------
    # Commands
    # --------

    @command(
    )
    @DebugIt()
    def start_acquisition(self):
        # PROTECTED REGION ID(StatsCrosslet.start_acquisition) ENABLED START #
        """
        Start the data acquisition of the station`s crosslet stats.

        :param argin: 'DevULong'

        :return:None
        """
        if self.is_start_acquisition_allowed() is True:
            self.data_acquisition_is_active = True
            self.set_state(DevState.RUNNING)
        # PROTECTED REGION END #    //  StatsCrosslet.start_acquisition

    def is_start_acquisition_allowed(self):
        # PROTECTED REGION ID(StatsCrosslet.is_start_acquisition_allowed) ENABLED START #
        return self.get_state() not in [DevState.ON,DevState.OFF,DevState.STANDBY,DevState.RUNNING]
        # PROTECTED REGION END #    //  StatsCrosslet.is_start_acquisition_allowed

    @command(
    )
    @DebugIt()
    def stop_acquisition(self):
        # PROTECTED REGION ID(StatsCrosslet.stop_acquisition) ENABLED START #
        """
        Stop the data acquisition.

        :return:None
        """
        if self.is_stop_acquisition_allowed() is True:
            self.set_state(DevState.ON)
            self.data_acquisition_is_active = False
        # PROTECTED REGION END #    //  StatsCrosslet.stop_acquisition

    def is_stop_acquisition_allowed(self):
        # PROTECTED REGION ID(StatsCrosslet.is_stop_acquisition_allowed) ENABLED START #
        return self.get_state() not in [DevState.ON,DevState.INIT,DevState.RUNNING]
        # PROTECTED REGION END #    //  StatsCrosslet.is_stop_acquisition_allowed

    @command(
    )
    @DebugIt()
    def On(self):
        # PROTECTED REGION ID(StatsCrosslet.On) ENABLED START #
        """

        :return:None
        """
        pass
        # PROTECTED REGION END #    //  StatsCrosslet.On

    @command(
    )
    @DebugIt()
    def Local(self):
        # PROTECTED REGION ID(StatsCrosslet.Local) ENABLED START #
        """

        :return:None
        """
        pass
        # PROTECTED REGION END #    //  StatsCrosslet.Local

    @command(
    )
    @DebugIt()
    def Remote(self):
        # PROTECTED REGION ID(StatsCrosslet.Remote) ENABLED START #
        """

        :return:None
        """
        pass
        # PROTECTED REGION END #    //  StatsCrosslet.Remote

    @command(
    )
    @DebugIt()
    def Reset(self):
        # PROTECTED REGION ID(StatsCrosslet.Reset) ENABLED START #
        """

        :return:None
        """
        pass
        # PROTECTED REGION END #    //  StatsCrosslet.Reset

    @command(
    )
    @DebugIt()
    def Off(self):
        # PROTECTED REGION ID(StatsCrosslet.Off) ENABLED START #
        """

        :return:None
        """
        pass
        # PROTECTED REGION END #    //  StatsCrosslet.Off

    @command(
    )
    @DebugIt()
    def Init(self):
        # PROTECTED REGION ID(StatsCrosslet.Init) ENABLED START #
        """

        :return:None
        """
        pass
        # PROTECTED REGION END #    //  StatsCrosslet.Init

# ----------
# Run server
# ----------


def main(args=None, **kwargs):
    """Main function of the StatsCrosslet module."""
    # PROTECTED REGION ID(StatsCrosslet.main) ENABLED START #
    return run((StatsCrosslet,), args=args, **kwargs)
    # PROTECTED REGION END #    //  StatsCrosslet.main


if __name__ == '__main__':
    main()
